// Dixit is Copyright (C) 2004-2013 Octavian Procopiuc.
// Homepage: http://dixit.sourceforge.net
// Contact:  Octavian Procopiuc <oprocopiuc@gmail.com>
//
// Use of this source code is governed by the GPLv3
// license that can be found in the LICENSE.GPL3 file,
// or at <http://www.gnu.org/licenses/gpl.html>.
//
// FILENAME:   mainwindow.cpp
// REVISION:   3
// COMMENTARY: Implements Qt4 MainWindow class

#include "mainwindow.h"
#include "dixit.h"
#include <zlib.h>

// MainWindow's constructor
MainWindow::MainWindow(void)
  : m_activeConnection(false)
  , m_autoUpdate(false)
  , m_logInfo(false)
  , m_updateError(UPDATE_ERR_NO_ERROR)
  , m_defaultDictionary(0)
  , m_lastUpdateTime(0)
  , m_lastUpdateTimePending(0)
  , m_newDefCount(0)
  , m_newDefCountPending(0)
  , m_totalDefinitions(0)
  , m_currentWord("")
  , m_html("")
  , m_lstAllWords()
  , m_lstBackwardWords()
  , m_lstForwardWords()
  , m_lstNewDefinitions()
  , m_connectionTimer(NULL)
  , w_dictionaryComboBox(NULL)
  , w_searchBox(NULL)
  , w_historyListBox(NULL)
  , w_browser(NULL)
  , w_toolBar(NULL)
  , a_next(NULL)
  , a_prev(NULL)
  , m_http(NULL)
  , m_dbHandle(NULL)
  , m_pThread(NULL) {
  CreateStatusBar();
  CreateMainWindow();
  CreateToolBar();
  CreateIndependentActions();
  LoadSettings();
  DbStartIndexing();
  return;
}

// MainWindow's destructor. Frees any allocated resources,
// no need to delete child widgets, Qt does it all
MainWindow::~MainWindow() {
  if (m_pThread) {
    m_pThread->wait();
    delete m_pThread;
  }

  if (m_http) {
    delete m_http;
  }

  SaveSettings();
  return;
}

// Called from the constructor.
void MainWindow::CreateStatusBar(void) {
  statusBar()->showMessage(QString::fromUtf8("Dicționar explicativ "
                           "al limbii române"));
  return;
}

// Called from the constructor.
void MainWindow::CreateMainWindow(void) {
  if (QFile::exists(USER_NEW_IMAGES_PATH + "carte.png")) {
    setWindowIcon(QIcon(USER_NEW_IMAGES_PATH + "carte.png"));

  } else {
    setWindowIcon(QIcon(":/images/book.png"));
  }

  setWindowTitle("Dixit");

  QSplitter *splitter = new QSplitter(this);
  splitter->setOrientation(Qt::Horizontal);

  w_browser = new QTextBrowser(splitter);
  w_browser->setMinimumSize(QSize(200, 0));
  w_browser->setAcceptRichText(true);
  w_browser->setAlignment(Qt::AlignJustify);

  // Set link color to dark blue, etc.
  QPalette pal = w_browser->palette();
  pal.setColor(QPalette::Active,   QPalette::Link, QColor("#003399"));
  pal.setColor(QPalette::Active,   QPalette::Text, QColor("#191919"));
  pal.setColor(QPalette::Inactive, QPalette::Link, QColor("#003399"));
  pal.setColor(QPalette::Inactive, QPalette::Text, QColor("#191919"));
  pal.setColor(QPalette::Disabled, QPalette::Link, QColor("#3366cc"));
  pal.setColor(QPalette::Disabled, QPalette::Text, QColor("#555555"));
  w_browser->setPalette(pal);

  connect(w_browser, SIGNAL(anchorClicked(const QUrl &)),
          this, SLOT(LinkClicked(const QUrl &)));
  connect(w_browser, SIGNAL(highlighted(const QString &)),
          (const QObject *) statusBar(), SLOT(message(const QString &)));
  splitter->addWidget(w_browser);

  w_historyListBox = new QListWidget(splitter);
  w_historyListBox->setViewMode(QListView::ListMode);
  w_historyListBox->setSelectionMode(QAbstractItemView::SingleSelection);
  QSizePolicy sizePolicy((QSizePolicy::Policy)1, (QSizePolicy::Policy)7);
  sizePolicy.setHorizontalStretch(0);
  sizePolicy.setVerticalStretch(0);
  sizePolicy.setHeightForWidth(w_historyListBox->sizePolicy()
                               .hasHeightForWidth());
  w_historyListBox->setSizePolicy(sizePolicy);
  w_historyListBox->setMinimumSize(QSize(125, 100));
  w_historyListBox->setMaximumSize(QSize(125, 32767));
  connect(w_historyListBox, SIGNAL(currentTextChanged(const QString &)),
          this, SLOT(WordClickedInHistoryBox(const QString &)));
  splitter->addWidget(w_historyListBox);
  setCentralWidget(splitter);
  return;
}

// Called from the constructor.
void MainWindow::CreateToolBar(void) {
  w_toolBar = addToolBar(tr("Dixit"));
  w_toolBar->setEnabled(true);
  w_toolBar->setFixedWidth(730);
  QFont tbFont;
  tbFont.setFamily(tr("Arial"));
  tbFont.setPointSize(10);
  tbFont.setBold(false);
  tbFont.setItalic(false);
  tbFont.setUnderline(false);
  tbFont.setStrikeOut(false);
  w_toolBar->setFont(tbFont);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "inapoi.png")) {
    a_prev = new QAction(QIcon(USER_NEW_IMAGES_PATH + "inapoi.png"),
                         QString::fromUtf8("Înapoi"),
                         this);

  } else {
    a_prev = new QAction(QIcon(":/images/prev.png"),
                         QString::fromUtf8("Înapoi"),
                         this);
  }

  a_prev->setShortcut(tr("Alt+Left"));
  a_prev->setStatusTip(QString::fromUtf8("Înapoi (F11 sau Alt+Stânga)"));
  a_prev->setEnabled(false);
  connect(a_prev, SIGNAL(triggered()), this, SLOT(BackwardInHistory()));
  w_toolBar->addAction(a_prev);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "inainte.png")) {
    a_next = new QAction(QIcon(USER_NEW_IMAGES_PATH + "inainte.png"),
                         QString::fromUtf8("Înainte"),
                         this);

  } else {
    a_next = new QAction(QIcon(":/images/next.png"),
                         QString::fromUtf8("Înainte"),
                         this);
  }

  a_next->setShortcut(tr("Alt+Right"));
  a_next->setStatusTip(QString::fromUtf8("Înainte (F12 sau Alt+Dreapta)"));
  a_next->setEnabled(false);
  connect(a_next, SIGNAL(triggered()), this, SLOT(ForwardInHistory()));
  w_toolBar->addAction(a_next);
  w_toolBar->addSeparator();

  QLabel *searchLabel = new QLabel(QString::fromUtf8("Cu&vântul:"), w_toolBar);
  searchLabel->setMinimumSize(QSize(35, 0));
  QFont lbFont;
  lbFont.setBold(true);
  lbFont.setItalic(false);
  lbFont.setUnderline(false);
  lbFont.setStrikeOut(false);
  searchLabel->setFont(lbFont);

  w_searchBox = new QLineEdit(w_toolBar);
  w_searchBox->setMinimumSize(QSize(175, 0));
  w_searchBox->setMaxLength(50);
  w_searchBox->setFrame(true);

  // Allow only the following on the search line: letters, *, ?, |, -, space.
  QRegExp re("!{0,1}[a-zA-Z*?|^\\[\\]#\\- ]+");
  w_searchBox->setValidator(new QRegExpValidator(re, this));
  connect(w_searchBox, SIGNAL(returnPressed()), this, SLOT(Search()));

  // buddies
  searchLabel->setBuddy(w_searchBox);

  w_toolBar->addWidget(searchLabel);
  w_toolBar->addWidget(w_searchBox);

  w_dictionaryComboBox = new QComboBox;
  w_dictionaryComboBox->setEditable(false);
  w_dictionaryComboBox->setMaxVisibleItems(DICTIONARIES_COUNT);
  QFont tbf = w_toolBar->font();
  tbf.setWeight(QFont::Bold);
  w_dictionaryComboBox->setFont(tbf);

  for (int row = 0; row < DICTIONARIES_COUNT; row++) {
    w_dictionaryComboBox->insertItem(row,
                                     QString::fromUtf8(DICTIONARIES[row][1]));
  }

  connect(w_dictionaryComboBox, SIGNAL(activated(int)),
          this, SLOT(SetDefaultDictionary(int)));
  connect(w_dictionaryComboBox, SIGNAL(highlighted(int)),
          this, SLOT(ShowDictionary(int)));
  w_toolBar->addWidget(w_dictionaryComboBox);

  QAction *action;

  if (QFile::exists(USER_NEW_IMAGES_PATH + "cauta.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "cauta.png"),
                         QString::fromUtf8("Caută"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/find.png"),
                         QString::fromUtf8("Caută"),
                         this);
  }

  action->setShortcut(tr("Alt+C"));
  action->setStatusTip(QString::fromUtf8("Caută (F3 sau Alt+C)"));
  connect(action, SIGNAL(triggered()), this, SLOT(Search()));
  w_toolBar->addAction(action);
  w_toolBar->addSeparator();

  if (QFile::exists(USER_NEW_IMAGES_PATH + "actualizeaza.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "actualizeaza.png"),
                         QString::fromUtf8("Actualizează lista de definiții"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/update.png"),
                         QString::fromUtf8("Actualizează lista de definiții"),
                         this);
  }

  action->setShortcut(tr("Alt+A"));
  action->setStatusTip(QString::fromUtf8("Actualizează lista de "
                                         "definiții (F5 sau Alt+A)"));
  connect(action, SIGNAL(triggered()), this, SLOT(Update()));
  w_toolBar->addAction(action);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "tipareste.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "tipareste.png"),
                         QString::fromUtf8("Tipărește"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/print.png"),
                         QString::fromUtf8("Tipărește"),
                         this);
  }

  action->setShortcut(tr("Alt+P"));
  action->setStatusTip(QString::fromUtf8("Tipărește pagina curentă "
                                         "(F2 sau Alt+P)"));
  connect(action, SIGNAL(triggered()), this, SLOT(Print()));
  w_toolBar->addAction(action);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "instructiuni.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "instructiuni.png"),
                         QString::fromUtf8("Instrucțiuni"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/help.png"),
                         QString::fromUtf8("Instrucțiuni"),
                         this);
  }

  action->setShortcut(tr("Alt+I"));
  action->setStatusTip(QString::fromUtf8("Instrucțiuni de folosire "
                                         "(F1 sau Alt+I)"));
  connect(action, SIGNAL(triggered()), this, SLOT(Help()));
  w_toolBar->addAction(action);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "iesire.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "iesire.png"),
                         QString::fromUtf8("Ieșire"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/exit.png"),
                         QString::fromUtf8("Ieșire"),
                         this);
  }

  action->setShortcut(tr("Alt+Q"));
  action->setStatusTip(QString::fromUtf8("Ieșire (F10, sau Alt+F4, "
                                         "sau Alt+Q)"));
  connect(action, SIGNAL(triggered()), this, SLOT(close()));
  w_toolBar->addAction(action);
  w_toolBar->addSeparator();
  return;
}

// Called from the constructor.
void MainWindow::CreateIndependentActions(void) {
  QAction *action;
  action = new QAction(this);
  action->setShortcut(tr("Esc"));
  connect(action, SIGNAL(triggered()), this, SLOT(AbortUpdate()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("Alt+F"));
  connect(action, SIGNAL(triggered()), this, SLOT(ChangeFont()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("Alt++"));
  connect(action, SIGNAL(triggered()), this, SLOT(MakeFontBigger()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("Alt+-"));
  connect(action, SIGNAL(triggered()), this, SLOT(MakeFontSmaller()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F1"));
  connect(action, SIGNAL(triggered()), this, SLOT(Help()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F2"));
  connect(action, SIGNAL(triggered()), this, SLOT(Print()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F3"));
  connect(action, SIGNAL(triggered()), this, SLOT(Search()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F5"));
  connect(action, SIGNAL(triggered()), this, SLOT(Update()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F6"));
  connect(action, SIGNAL(triggered()), this, SLOT(ChangeFont()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F7"));
  connect(action, SIGNAL(triggered()), this, SLOT(MakeFontBigger()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F8"));
  connect(action, SIGNAL(triggered()), this, SLOT(MakeFontSmaller()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F10"));
  connect(action, SIGNAL(triggered()), this, SLOT(close()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F11"));
  connect(action, SIGNAL(triggered()), this, SLOT(BackwardInHistory()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("F12"));
  connect(action, SIGNAL(triggered()), this, SLOT(ForwardInHistory()));
  addAction(action);

  m_http = new QHttp(this);
  connect(m_http, SIGNAL(responseHeaderReceived(const QHttpResponseHeader &)),
          this, SLOT(CheckHttpResponseHeader(const QHttpResponseHeader &)));
  connect(m_http, SIGNAL(readyRead(const QHttpResponseHeader &)),
          this, SLOT(GetDexonlineData(const QHttpResponseHeader &)));
  connect(m_http, SIGNAL(done(bool)), this, SLOT(CloseConnection(bool)));
  connect(&m_connectionTimer, SIGNAL(timeout()),
          this, SLOT(ConnectionTimeout()));

  return;
}

// Called from the constructor.
void MainWindow::LoadSettings(void) {
  QSettings settings(APP_GROUP, APP_KEY);
  QPoint pos = settings.value("WindowPos", QPoint(100, 100)).toPoint();
  QSize size = settings.value("WindowSize", QSize(770, 500)).toSize();
  m_autoUpdate = settings.value("AutoUpdate", bool(false)).toBool();
  m_logInfo = settings.value("LogInfo", bool(false)).toBool();
  m_defaultDictionary = settings.value("DefaultSource", int(0)).toInt();
  // keep the default font family, if none is chosen by the user
  QString fontFamily = font().family();
  fontFamily = settings.value("FontFamily", QString(fontFamily)).toString();
  int fontPointSize = settings.value("FontPointSize", int(12)).toInt();

  w_dictionaryComboBox->setCurrentIndex(m_defaultDictionary);
  resize(size);
  move(pos);

  if (fontPointSize > 16) {
    fontPointSize = 16;

  } else if (fontPointSize < 8) {
    fontPointSize = 8;
  }

  QFont cf = font();
  cf.setFamily(fontFamily);
  cf.setPointSize(fontPointSize);
  cf.setStyleStrategy(QFont::PreferAntialias);
  setFont(cf);
  return;
}

// Called from the destructor.
void MainWindow::SaveSettings(void) {
  QSettings settings(APP_GROUP, APP_KEY);
  settings.setValue("WindowPos", pos());
  settings.setValue("WindowSize", size());
  settings.setValue("AutoUpdate", m_autoUpdate);
  settings.setValue("LogInfo", m_logInfo);
  settings.setValue("DefaultSource", m_defaultDictionary);
  settings.setValue("FontFamily", font().family());
  settings.setValue("FontPointSize", font().pointSize());
  return;
}

QString MainWindow::LoadFile(const QString &fileName) {
  QFile file(fileName);
  file.open(QIODevice::ReadOnly);
  QByteArray htmlData = file.readAll();
  file.close();
  return QString::fromUtf8(htmlData.constData());
}

// This slot is linked to the print button.
void MainWindow::Print(void) {
#ifndef QT_NO_PRINTER
  statusBar()->showMessage(QString::fromUtf8("Tipărește"));
  QTextDocument *document = w_browser->document();
  QPrinter printer;
  QPrintDialog *dlg = new QPrintDialog(&printer, this);

  if (dlg->exec() != QDialog::Accepted) {
    return;
  }

  document->print(&printer);
  statusBar()->showMessage(QString::fromUtf8("Tipărește"), 2000);
#else
  statusBar()->showMessage(QString::fromUtf8("Programul nu poate tipări"),
                           10000);
  ClearBrowser();
  SetBrowserHtml(QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" "
                                   "width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
                                   "<font color='#bb3300'><b>Programul nu "
                                   "poate tipări</b></font></td></tr><tr><td>"
                                   "<font color='#bb3300'>Biblioteca Qt4 nu a "
                                   "fost compilată cu suport pentru tipărire "
                                   "pentru platforma dumneavoastră ("
                                   "QT_NO_PRINTER <u>was defined</u>).</font>"
                                   "</td></tr></table>"));
#endif
  return;
}

void MainWindow::ProcessWord(const QString &word) {
  if (word.isEmpty()) {
    return;
  }

  QString lWord(word);
  lWord = lWord.toLower();

  if (m_currentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_currentWord);
  }

  InsertInSearchBox(lWord);
  FindDefinitions(lWord);
  return;
}

// This slot is linked to the help button.
void MainWindow::Help(void) {
  ProcessWord("!instructiuni");
  return;
}

// This slot is linked to w_dictionaryComboBox selection action
void MainWindow::SetDefaultDictionary(int id) {
  m_defaultDictionary = id;
  w_dictionaryComboBox->setCurrentIndex(id);
  return;
}

// This slot is linked to w_dictionaryComboBox selection action
void MainWindow::ShowDictionary(int id) {
  statusBar()->showMessage(QString::fromUtf8(DICTIONARIES[id][2]));
  return;
}

// This slot is linked to the input text box and search button.
void MainWindow::Search(void) {
  QString word = w_searchBox->text();

  if (word.isEmpty()) {
    return;
  }

  word = word.toLower();

  if (m_currentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_currentWord);
    m_lstForwardWords.clear();
  }

  FindDefinitions(word);
  return;
}

// This slot is linked to the history box item click.
void MainWindow::WordClickedInHistoryBox(const QString &word) {
  if (word.isEmpty() || word == m_currentWord) {
    return;
  }

  if (m_currentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_currentWord);
    m_lstForwardWords.clear();
  }

  InsertInSearchBox(word);
  FindDefinitions(word);
  return;
}

// This slot is linked to the w_browser click link.
void MainWindow::LinkClicked(const QUrl &link) {
  QString word = link.toString();
  ProcessWord(word);
  return;
}

// This slot is linked to the backward button.
void MainWindow::BackwardInHistory(void) {
  if (m_lstBackwardWords.isEmpty()) {
    a_prev->setEnabled(false);
    return;
  }

  m_lstForwardWords.push_back(m_currentWord);
  InsertInSearchBox(m_lstBackwardWords.last());
  FindDefinitions(m_lstBackwardWords.last());
  m_lstBackwardWords.pop_back();

  if (m_lstBackwardWords.isEmpty()) {
    a_prev->setEnabled(false);

  } else {
    a_prev->setEnabled(true);
  }

  return;
}

// This slot is linked to the forward button.
void MainWindow::ForwardInHistory(void) {
  if (m_lstForwardWords.isEmpty()) {
    a_next->setEnabled(false);
    return;
  }

  m_lstBackwardWords.push_back(m_currentWord);
  InsertInSearchBox(m_lstForwardWords.last());
  FindDefinitions(m_lstForwardWords.last());
  m_lstForwardWords.pop_back();

  if (m_lstForwardWords.isEmpty()) {
    a_next->setEnabled(false);

  } else {
    a_next->setEnabled(true);
  }

  return;
}

void MainWindow::FindDefinitions(const QString &word) {
  if (word.isEmpty()) {
    return;
  }

  QListWidgetItem *item;
  QList<QListWidgetItem *> foundItems;

  // Limit the history.
  if (w_historyListBox->count() >= MAX_HISTORY_SIZE) {
    item = w_historyListBox->takeItem(w_historyListBox->count() - 1);

    // Items removed from a list widget need to be deleted manually.
    if (item) {
      delete item;
    }
  }

  w_historyListBox->clearSelection();
  foundItems = w_historyListBox->findItems(word, Qt::MatchExactly);

  if (foundItems.count() > 0) {  // if item is found, highlight
    item = foundItems.front();
    w_historyListBox->setItemSelected(item, true);

  } else {  // else insert word at top of history list.
    w_historyListBox->insertItem(0, word);
    item = w_historyListBox->item(0);
    w_historyListBox->setItemSelected(item, true);
  }

  if (m_lstForwardWords.isEmpty()) {
    a_next->setEnabled(false);

  } else {
    a_next->setEnabled(true);
  }

  if (m_lstBackwardWords.isEmpty()) {
    a_prev->setEnabled(false);

  } else {
    a_prev->setEnabled(true);
  }

  // process all commands
  if ('!' == word[0]) {
    // Check for commands implying QActions.
    if ("!maimare" == word) {
      MakeFontBigger();
      return;

    } else if ("!maimic" == word) {
      MakeFontSmaller();
      return;
    }

    // Find the anchor for *.html#anchor browser navigation, if applicable
    QString lWord(word);
    QString lanchor = lWord.section('#', 1);
    lWord = lWord.section('#', 0, 0);
    ClearBrowser();
    // Display html page
    SetBrowserHtml(ProcessCommands(lWord));

    if (!lanchor.isEmpty()) {
      w_browser->scrollToAnchor(lanchor);
    }

    w_searchBox->selectAll();
    return;
  }

  // search for word definitions in database
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
  m_currentWord = word;
  ClearBrowser();

  if (m_pThread) {
    m_pThread->wait();
  }

  SetBrowserHtml(DbResults(word));
  QApplication::restoreOverrideCursor();

  SetFocusToSearchBox();
  return;
}

void MainWindow::SetFocusToSearchBox(void) {
  w_searchBox->setFocus();
  w_searchBox->selectAll();
  return;
}

// Insert word in the search box
void MainWindow::InsertInSearchBox(const QString &word) {
  w_searchBox->setText(word);
  return;
}

void MainWindow::ClearBrowser(void) {
  m_html = "";
  w_browser->clear();
  return;
}

void MainWindow::SetBrowserHtml(const QString &str,
                                bool prs) {
  w_browser->setHtml(m_html + str);

  if (prs) {  // persist.
    m_html += str;
  }

  return;
}

// This slot is linked to the update button.
void MainWindow::Update(void) {
  w_searchBox->clear();
  w_historyListBox->clearSelection();
  ClearBrowser();

  if (m_currentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_currentWord);
    m_lstForwardWords.clear();
    m_currentWord = "";
  }

  GenericUpdate();
  return;
}

// Slot linked to m_connectionTimer.
void MainWindow::AbortUpdate(void) {
  if (QHttp::Unconnected != m_http->state()
      &&  QHttp::Closing != m_http->state()) {
    m_updateError = UPDATE_ERR_ABORTED;
    m_http->abort();
  }

  return;
}

// Slot linked to the m_connectionTimer.
void MainWindow::ConnectionTimeout(void) {
  m_updateError = UPDATE_ERR_TIMEOUT;
  m_http->abort();
  return;
}

// Slot linked to QHttp::done.
void MainWindow::CloseConnection(bool error) {
  m_connectionTimer.stop();

  // Close the database.
  DbClose();

  if (UPDATE_ERR_NO_ERROR == m_updateError) {
    error = false;
  }

  if (!error && -1 == m_newDefCountPending) {
    m_updateError = UPDATE_ERR_CONTENT;
    error = true;
  }

  if (error) {
    // Treat some special cases.
    if (UPDATE_ERR_TIMEOUT == m_updateError
        || UPDATE_ERR_ABORTED == m_updateError) {
      // "Minor" error. Recover gracefully below.
      error = false;

    } else if (!m_updateError) {
      // Got no chance to set m_updateError, probably due to a network error.
      m_updateError = UPDATE_ERR_NETWORK;
    }
  }

  if (UPDATE_ERR_DB_READ != m_updateError
      &&  UPDATE_ERR_DB_WRITE != m_updateError) {
    m_lastUpdateTime = m_lastUpdateTimePending;

    if (DbOpenWrite()) {
      QString luts = QString("%1").arg(m_lastUpdateTime);

      // Store the new update time in the database.
      if (!DbPut(LAST_UPDATE_TIME_KEY, luts.toAscii().data())) {
        LogInfo("Error: New update time was not saved.");
        m_updateError = UPDATE_ERR_DB_WRITE;
        error = true;
      }

      DbClose();
    }
  }

  if (!error) {
    QString wc =
      QString::fromUtf8(" Terminat. %1 definiții noi sau actualizate.<br>")
      .arg(m_newDefCount);

    if (0 != m_newDefCount) {
      wc +=
        QString::fromUtf8("Cele mai noi definiții pot fi găsite "
                          "<a href=\"!nou\">aici</a> (maxim %1).<br>")
        .arg(MAX_NEW_DEFS_TO_DISPLAY);
    }

    if (m_autoUpdate && 0 == w_historyListBox->count()) {
      wc +=
        QString::fromUtf8("<br>Baza de date se actualizează automat la "
                          "pornirea programului. Pentru a schimba această "
                          "preferință apăsați <a href=\"!lazimanual\">aici"
                          "</a>. Actualizarea se poate face în orice moment "
                          "folosind butonul de actualizare.<br>");
    }

    SetBrowserHtml(wc);

    if (UPDATE_ERR_TIMEOUT == m_updateError
        || UPDATE_ERR_NOT_DONE == m_updateError
        || UPDATE_ERR_ABORTED == m_updateError) {
      LogInfo("Error: Connection timeout or abort. Download cut short.");
      wc =
        QString::fromUtf8("<font color=\"#993300\">Legătura s-a întrerupt "
                          "prematur. Apăsați butonul de actualizare pentru a "
                          "descărca restul de definiții.</font><br>");
      LogInfo(QString("Info: Last update time %1").arg(m_lastUpdateTime));
      SetBrowserHtml(wc);

    } else {
      LogInfo("Info: Finished all data from server.");
    }

  } else {  // if an error
    LogInfo(QString::fromUtf8("Error: %1.")
            .arg(UPDATE_ERRORS_STRING[m_updateError]));
    QString wc = QString::fromUtf8(
                   "<font color='#993300'>Eroare (%1).</font><br> Data ultimei "
                   "actualizări (sau a ultimului cuvânt actualizat): %2.<br>"
                 )
                 .arg(QString::fromUtf8(UPDATE_ERRORS_STRING[m_updateError]))
                 .arg(FormatRoDate(m_lastUpdateTime));

    if (UPDATE_ERR_DB_READ == m_updateError
        || UPDATE_ERR_DB_WRITE == m_updateError) {
      wc += QString::fromUtf8("Eroare în baza de date.");
    }

    SetBrowserHtml(wc);
  }

  // Build an in-memory word list.
  if (0 != m_newDefCount) {
    m_pThread->start();
  }

  w_browser->setEnabled(true);
  w_historyListBox->setEnabled(true);
  w_toolBar->setEnabled(true);
  SetFocusToSearchBox();
  return;
}

// This slot is linked to the QHttp::responseHeaderReceived signal.
void MainWindow::CheckHttpResponseHeader(const QHttpResponseHeader &rh) {
  if (200 != rh.statusCode()) {
    LogInfo(QString("Error: HTTP server returned status code %1 %2.")
            .arg(rh.statusCode())
            .arg(rh.reasonPhrase()));
    m_updateError = UPDATE_ERR_SERVER;
    m_http->abort();

  } else if (!(rh.contentType().contains("gzip")
               || rh.value(QString::fromAscii("content-encoding"))
               .contains("gzip"))) {
    LogInfo(QString("Error: Unknown content type received from server: %1")
            .arg(rh.contentType()));
    m_updateError = UPDATE_ERR_CONTENT;
    m_http->abort();
  }

  return;
}

// This slot is linked to the QHttp::readyRead signal.
void MainWindow::GetDexonlineData(const QHttpResponseHeader &rh) {
  m_connectionTimer.setSingleShot(true);
  m_connectionTimer.start(20 * 1000);

  static QByteArray xmlData;
  static bool gzipped = false;

  if (m_activeConnection) {
    m_newDefCount = 0;
    m_newDefCountPending = -1;

    if (200 != rh.statusCode()) {
      LogInfo(QString("Error: HTTP server returned status code %1 %2.")
              .arg(rh.statusCode())
              .arg(rh.reasonPhrase()));
      m_updateError = UPDATE_ERR_SERVER;
      m_http->abort();
      return;
    }

    // Initialize the zlib inflate structures, if necessary.
    if (rh.contentType().contains("gzip")
        || rh.value(QString::fromAscii("content-encoding")).contains("gzip")) {
      gzipped = true;
      ProcessGzipData(&xmlData, NULL, 0, true);

    } else {
      LogInfo(QString("Error: Unknown content type received from server: %1")
              .arg(rh.contentType()));
      m_updateError = UPDATE_ERR_CONTENT;
      m_http->abort();
      return;
    }

    xmlData.truncate(0);
    m_activeConnection = false;
  }

  int dataLen = xmlData.length();

  if (dataLen < 1000000) {
    int baSz = m_http->bytesAvailable();

    if (!baSz) {
      LogInfo("Error: GetDexonlineData called, but no XML data is available.");
      return;
    }

    // Don't read more than this at a time.
    if (baSz > 1000000) {
      baSz = 1000000;
    }

    char *buf = new char[baSz + 1];
    int bufSz = m_http->read(buf, baSz);

    if (-1 == bufSz) {
      LogInfo("Error: m_http->read(buf, baSz) returned -1).");
      m_updateError = UPDATE_ERR_STREAM;
      m_http->abort();
      delete [] buf;
      return;
    }

    // Inflate, if needed.
    if (gzipped) {
      ProcessGzipData(&xmlData, buf, bufSz, false);

    } else {
#if QT_VERSION >= 0x040500
      xmlData.append(buf, bufSz);
#else
      AppendToByteArray(&xmlData, buf, bufSz);
#endif
    }

    delete [] buf;
  }

  // Process XML data when 900k chars are available.
  dataLen = xmlData.length();
  bool bytesAvailable = (m_http->bytesAvailable() ? true : false);

  if (dataLen >= 900000 || !(bytesAvailable)) {
    bool abortHttp = false;
    int lastIdx = DbSaveXMLData(xmlData, &m_updateError, &abortHttp);

    if (abortHttp) {
      m_http->abort();
      return;
    }

    // If anything processed, discard it.
    if (0 != lastIdx) {
      xmlData = xmlData.mid(lastIdx, dataLen);
    }

    QString wc =
      QString::fromUtf8(" <b>%1/%2</b> definiții noi sau actualizate "
                        "(apăsați <i>Esc</i> pentru a întrerupe).")
      .arg(m_newDefCount)
      .arg(m_newDefCountPending);
    SetBrowserHtml(wc, false);
  }

  if (bytesAvailable) {
    emit MoreData(rh);
  }

  return;
}

// Start a connection with the server.
void MainWindow::GenericUpdate(void) {
  if (m_pThread) {
    m_pThread->wait();
  }

  QString buf =
    QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" width="
                      "\"100%\"><tr><td bgcolor=\"#eeeeee\"><b>Actualizare</b>"
                      "</td></tr><tr><td>Programul încearcă să contacteze "
                      "serverul %1... ")
    .arg(DEFAULT_SERVER_NAME);
  SetBrowserHtml(buf);
  w_searchBox->clearFocus();
  w_toolBar->setEnabled(false);
  w_browser->setEnabled(false);
  w_historyListBox->setEnabled(false);

  m_updateError = UPDATE_ERR_NO_ERROR;
  m_lastUpdateTimePending = m_lastUpdateTime;

  buf = HTTP_GET_REQUEST;
  QHttpRequestHeader header("GET", buf.arg(m_lastUpdateTime));
  header.setValue("Host", DEFAULT_SERVER_NAME);

  buf = USER_AGENT;
  header.setValue("User-Agent", buf.arg(DIXIT_VERSION));
  header.setValue("Accept-Encoding", "gzip");
  header.setValue("Connection", "Keep-Alive");

  // Timeout after 20 seconds.
  m_connectionTimer.setSingleShot(true);
  m_connectionTimer.start(20 * 1000);
  m_activeConnection = true;
  m_newDefCount = 0;
  m_newDefCountPending = -1;

  m_http->setHost(DEFAULT_SERVER_NAME);
  m_http->request(header);
  return;
}

// Implements the Levenshtein distance algorithm found at
// http://www.merriampark.com/ldcpp.htm. If 0 == maxDist, return the true
// Levenshtein (edit) distance. Otherwise, return the minimum between maxDist
// and the edit distance. If source or target are longer than 50, returns -1.
int MainWindow::LevenshteinDistance(const QString &source,
                                    const QString &target,
                                    int maxDist) {
  char n, m;
  int l;

  l = source.length();

  if (l > 50) {
    return -1;
  }

  n = l;

  l = target.length();

  if (l > 50) {
    return -1;
  }

  m = l;

  if (0 == maxDist) {
    maxDist = qMax(m, n);
  }

  if (0 == n) {
    return qMin(static_cast<int>(m), maxDist);
  }

  if (0 == m) {
    return qMin(static_cast<int>(n), maxDist);
  }

  // Store the min. value on each column, so that,
  // if it reaches maxDist, we break early.
  char minColVal;

  char matrix[51][51];

  char j;
  char i;
  char cell;

  for (j = 0; j <= m; j++) {
    matrix[0][static_cast<int>(j)] = j;
  }

  for (i = 1; i <= n; i++) {
    minColVal = qMax(m, i);
    matrix[static_cast<int>(i)][0] = i;

    const QChar s_i = source[i - 1];

    for (j = 1; j <= m; j++) {
      const QChar t_j = target[j - 1];
      char  cost = (s_i == t_j ? 0 : 1);
      const char above = matrix[static_cast<int>(i) - 1][static_cast<int>(j)];
      const char left = matrix[static_cast<int>(i)][static_cast<int>(j) - 1];
      const char diag =
        matrix[static_cast<int>(i) - 1][static_cast<int>(j) - 1];
      cell = qMin(above + 1, qMin(left + 1, diag + cost));

      // Cover transposition, in addition to deletion, insertion and
      // substitution. This step is taken from Hal Berghel and David Roach:
      // "An Extension of Ukkonen's Enhanced Dynamic Programming ASM Algorithm"
      // http://www.berghel.net/publications/asm/asm.php

      if (i > 2 && j > 2) {
        char trans =
          matrix[static_cast<int>(i) - 2][static_cast<int>(j) - 2] + 1;

        if (source[i - 2] != t_j) {
          trans++;
        }

        if (s_i != target[j - 2]) {
          trans++;
        }

        if (cell > trans) {
          cell = trans;
        }
      }

      minColVal = qMin(minColVal, cell);
      matrix[static_cast<int>(i)][static_cast<int>(j)] = cell;
    }

    if (minColVal >= maxDist) {
      break;
    }
  }

  if (i == n + 1) {
    return static_cast<int>(matrix[static_cast<int>(n)][static_cast<int>(m)]);

  } else {
    return maxDist;
  }
}

void MainWindow::LogInfo(const QString &msg) {
  if (m_logInfo) {
    if (m_pThread) {
      m_pThread->wait();
    }

    qWarning("%s", msg.toAscii().data());
  }

  return;
}

QString MainWindow::FormatRoDate(int seconds,
                                 bool hour) {
  if (0 == seconds) {
    return QString::fromUtf8("Niciodată");
  }

  QDateTime qtTime = QDateTime::fromTime_t(seconds);
  return QString("%1")
         .arg(qtTime.toString(hour ? "d.M.yyyy, h:mm" : "d.M.yyyy"));
}

QString MainWindow::ProcessCommands(const QString &cmd) {
  QString result;

  if (cmd.startsWith("!surs") || cmd.startsWith("!dex") || "!dar" == cmd
      || "!der" == cmd  || "!dge" == cmd || "!dlra" == cmd || "!dlrlv" == cmd
      || "!dlrlc" == cmd || "!dor" == cmd || "!dlrc" == cmd || "!dlrm" == cmd
      || "!dmlr" == cmd || "!doom" == cmd || "!mda" == cmd || "!mdn" == cmd
      || "!neoficial" == cmd || cmd.startsWith("!anton")) {
    result = LoadFile(":/html/sources.html");

  } else if ("!info" == cmd || "!despre" == cmd) {
    result =
      QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" width="
                        "\"100%\"><tr><td bgcolor=\"#eeeeee\"><b>Dixit</b></td>"
                        "</tr><tr><td><p align=\"justify\">Dixit permite "
                        "căutări de cuvinte <i>offline</i> folosind "
                        "definițiile adunate de proiectul <i>DEX Online "
                        "&lt;http://dexonline.ro&gt;</i>. Baza de date locală "
                        "conține %1 definiții culese din diverse <a href="
                        "\"!surse\">surse tipărite sau orale</a>. Data ultimei "
                        "actualizări: %2.<br></p></td></tr>")
      .arg(m_totalDefinitions)
      .arg(FormatRoDate(m_lastUpdateTime, true))

#if defined(Q_CC_MSVC)
      + LoadFile(":/html/lastinfow.html");
#else
      + LoadFile(":/html/lastinfo.html");
#endif

  } else if ("!instructiuni" == cmd || "!ajutor" == cmd || "!help" == cmd) {
    result = LoadFile(":/html/help.html")
             + QString::fromUtf8("<p align=\"justify\">Baza de date conținând "
                                 "definițiile este păstrată local și %1 se "
                                 "actualizează la pornirea programului%2. "
                                 "Pentru a schimba această preferință apăsați "
                                 "<a href=\"!lazi%3\">aici</a>. Actualizarea "
                                 "se poate face în orice moment folosind "
                                 "butonul de actualizare.</p>")
             .arg(m_autoUpdate ? "" : "nu")
             .arg(m_autoUpdate ? QString::fromUtf8(" dacă există o legătură "
                  "la Internet") : "")
             .arg(m_autoUpdate ? "manual" : "auto")

             + LoadFile(":/html/cmd.html")

             + QString::fromUtf8("<p align=\"justify\">Programul %1 afișează "
                                 "informațiile interne referitoare la execuția "
                                 "programului. Pentru a schimba această "
                                 "preferință apăsați <a href=\"!loginfo\">aici"
                                 "</a>.</p></td></tr></table>")
             .arg(m_logInfo ? "" : "nu");

  } else if (cmd.startsWith("!lazi")) {
    if (cmd.startsWith("!laziauto")) {
      m_autoUpdate = true;

    } else if (cmd.startsWith("!lazimanu")) {
      m_autoUpdate = false;
    }

    result =
      QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" width="
                        "\"100%\"><tr><td bgcolor=\"#eeeeee\"><b>Modul de "
                        "actualizare a fost schimbat</b></td></tr><tr><td>"
                        "Lista de definiții <b>%1 se actualizează automat</b> "
                        "la pornirea programului. Pentru a schimba această "
                        "preferință apăsați <a href=\"!lazi%2\">aici</a>."
                        "</td></tr></table>")
      .arg(m_autoUpdate ? "" : "nu")
      .arg(m_autoUpdate ? "manual" : "auto");

  } else if (cmd.startsWith("!abr") || cmd.startsWith("!presc")) {
    result = LoadFile(":/html/abrv.html");

  } else if ("!nou" == cmd || "!recent" == cmd) {
    result =
      QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" width="
                        "\"100%\"><tr><td bgcolor=\"#eeeeee\"><b>Cele mai noi "
                        "definiții (maxim %1)</b></td></tr>")
      .arg(MAX_NEW_DEFS_TO_DISPLAY);

    if (m_lstNewDefinitions.isEmpty()) {
      result +=
        QString::fromUtf8("<tr><td><p>Nimic nou. Data ultimei actualizări: %1."
                          "</p><br></td></tr></table>")
        .arg(FormatRoDate(m_lastUpdateTime));

    } else {
      result +=
        QString::fromUtf8("<tr><td align=\"right\"><font size=-1 color="
                          "'#686868'><a href=\"!abrevieri\">Abrevieri</a>"
                          "</font></td></tr></table>");
      int m = 0;

      for (QStringList::iterator it = m_lstNewDefinitions.begin();
           it != m_lstNewDefinitions.end();
           it++) {
        if (m > MAX_NEW_DEFS_TO_DISPLAY) {
          break;
        }

        result += HtmlizeBrowserEntry(*it);
        m++;
      }
    }

  } else if ("!gplro" == cmd) {
    result = LoadFile(":/html/gplro.html");

  } else if ("!gpl" == cmd) {
    result = LoadFile(":/html/gpl.html");

  } else if ("!dixit" == cmd) {
    result = LoadFile(":/html/dixit.html");

  } else if ("!pp" == cmd) {
    result = LoadFile(":/html/fp.html");

  } else if ("!loginfo" == cmd) {
    m_logInfo = !m_logInfo;
    result =
      QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" width="
                        "\"100%\"><tr><td bgcolor=\"#eeeeee\"><b>Informații "
                        "specifice execuției programului</b></td></tr><tr><td> "
                        "Informațiile specifice execuției programului <b>%1 "
                        "sunt afișate</b>. Pentru a schimba această preferință "
                        "apăsați <a href=\"!loginfo\">aici</a>.</td></tr>"
                        "</table>")
      .arg(m_logInfo ? "" : "nu");

  } else {
    // unknown command
    result =
      QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"2\" width="
                        "\"100%\"><tr><td bgcolor=\"#eeeeee\"><b>Programul nu "
                        "poate procesa comanda </b></td></tr><tr><td>Apăsați "
                        "<a href=\"!instructiuni\">aici</a> pentru a consulta "
                        "pagina de instrucțiuni.</td></tr></table>");
  }

  return result;
}

#if QT_VERSION < 0x040500
void MainWindow::AppendToByteArray(QByteArray *ba,
                                   const char *buf,
                                   int sz) {
  int oldsz = ba->size();
  ba->resize(oldsz + sz);
  memcpy(ba->begin() + oldsz, buf, sz);
  return;
}
#endif

// Process zData and append inflated result to iData.
void MainWindow::ProcessGzipData(QByteArray *iData,
                                 char *zData,
                                 int zDataSz,
                                 bool initOnly) {
  int error, len;
  bool done = false;
  char gzBuffer[65536];
  static z_stream gzStream;

  if (initOnly) {
    // Initialize zlib inflate.
    gzStream.zalloc = NULL;
    gzStream.zfree = NULL;
    gzStream.opaque = NULL;
    gzStream.next_in = gzStream.next_out = 0;
    gzStream.avail_in = gzStream.avail_out = 0;

    if (Z_OK != inflateInit2(&gzStream, 16 + 15)) {
      LogInfo("Error: Initializing gzip inflate needs zlib 1.2.x.");
    }

    return;
  }

  gzStream.next_out = reinterpret_cast<Bytef *>(gzBuffer);
  gzStream.avail_out = sizeof(gzBuffer);
  gzStream.next_in = reinterpret_cast<Bytef *>(zData);
  gzStream.avail_in = zDataSz;

  while (0 != gzStream.avail_in) {
    if (0 == gzStream.avail_out) {
#if QT_VERSION >= 0x040500
      iData->append(gzBuffer, sizeof(gzBuffer));
#else
      AppendToByteArray(iData, gzBuffer, sizeof(gzBuffer));
#endif
      gzStream.next_out = reinterpret_cast<Bytef *>(gzBuffer);
      gzStream.avail_out = sizeof(gzBuffer);
    }

    error = inflate(&gzStream, Z_NO_FLUSH);

    if (Z_STREAM_END == error) {
      break;

    } else if (Z_OK != error) {
      LogInfo("Error: The stream is not gzipped in ProcessGzipData "
              "or zlib version is too old (zlib 1.2.x needed).");
      exit(1);
    }
  }

  while (true) {
    len = sizeof(gzBuffer) - gzStream.avail_out;

    if (len > 0) {
#if QT_VERSION >= 0x040500
      iData->append(gzBuffer, len);
#else
      AppendToByteArray(iData, gzBuffer, len);
#endif
      gzStream.next_out  = reinterpret_cast<Bytef *>(gzBuffer);
      gzStream.avail_out = sizeof(gzBuffer);
    }

    if (done) {
      break;
    }

    error = inflate(&gzStream, Z_SYNC_FLUSH);

    if (0 == len  &&  Z_BUF_ERROR == error) {
      error = Z_OK;
    }

    done = (0 != gzStream.avail_out  ||  Z_STREAM_END == error);

    if (Z_OK != error  &&  Z_STREAM_END != error) {
      break;
    }
  }

  return;
}

// Replace i):    #, @, $, %, |, ^, _  with HTML tags
//         ii):   *, **, -, ""         with unicode
//         iii):  <                    with Din
QString MainWindow::HtmlizeRawDefinition(const char *rawDef) {
  bool inAbr = false, inBold = false, inItalic = false, inQuote = false;
  QChar h_i;
  int i = 0;
  int m, n;
  QString text, link, html_expr;
  QString html = QString::fromUtf8(rawDef);

  if (0 == html.length()) {
    return html;
  }

  do {
    h_i = html[i];

    if ('#' == h_i) {  // #abr.#
      if ('&' != html[i - 1]) {  // don't replace unicode chars (e.g. &#x2013;)
        html.replace(i, 1, (inAbr ? "</font>" : "<font color=DarkViolet>"));
        inAbr = !inAbr;
      }

    } else if ('@' == h_i) {  // @bold@
      if ('\\' != html[i - 1]) {  // don't replace escaped @ (\@)
        html.replace(i, 1, (inBold ? "</b>" : "<b>"));
        inBold = !inBold;
      }

    } else if ('$' == h_i) {  // $italic$
      if ('\\' != html[i - 1]) {  // don't replace escaped $ (\$)
        html.replace(i, 1, (inItalic ? "</i>" : "<i>"));
        inItalic = !inItalic;
      }

    } else if ('"' == h_i) {  // "quote"
      // avoid processing <a href="link">text</a> introduced for |text|link|
      if (!('=' == html[i - 1] || '>' == html[i + 1])) {
        html.replace(i, 1, (inQuote ? "&#8220;" : "&#8222;"));
        inQuote = !inQuote;
      }

    } else if ('%' == h_i) {  // %aerisit%
      if ('\\' != html[i - 1]) {  // don't replace escaped % (\%)
        m = html.indexOf("%", i + 1);  // position of the 2nd %

        if (-1 != m) {
          text = html.mid(i + 1, m - i - 1);
          n = text.length();

          if (n < 20) {  // In DE there are no escaped % (\%)
            html_expr = "";

            for (m = 0; m < n; m++) {
              if (m != n - 1) {
                html_expr += text[m] + " ";

              } else {
                html_expr += text[m];
              }
            }

            html.replace(i, n + 2, html_expr);
          }
        }
      }

    } else if ('|' == h_i) {  // |text|link|
      m = html.indexOf("|", i + 1);  // position of the 2nd bar
      n = html.indexOf("|", m + 1);  // position of the 3rd bar

      if (-1 != m  &&  -1 != n) {
        text = html.mid(i + 1, m - i - 1);
        link = html.mid(m + 1, n - m - 1);
        html_expr = "<a href=\"" + link + "\">" + text + "</a>";
        html.replace(i, text.length() + link.length() + 3, html_expr);
      }

    } else if ('^' == h_i) {  // <sup></sup>
      if ('{' != html[i + 1]) {
        text = html.mid(i + 1, 1);
        html_expr = "<sup>" + text + "</sup>";
        html.replace(i, 2, html_expr);

      } else {
        m = html.indexOf("}", i + 2);

        if (-1 != m) {
          text = html.mid(i + 2, m - i - 2);
          html_expr = "<sup>" + text + "</sup>";
          html.replace(i, text.length() + 3, html_expr);
        }
      }

    } else if ('_' == h_i) {  // <sub></sub>
      if ('{' != html[i + 1]) {
        text = html.mid(i + 1, 1);
        html_expr = "<sub>" + text + "</sub>";
        html.replace(i, 2, html_expr);

      } else {
        m = html.indexOf("}", i + 2);

        if (-1 != m) {
          text = html.mid(i + 2, m - i - 2);
          html_expr = "<sub>" + text + "</sub>";
          html.replace(i, text.length() + 3, html_expr);
        }
      }

    } else if ('*' == h_i) {  // put a diamond char
      if ('*' == html[i + 1]) {
        html.replace(i, 2, "&diams;");  // **

      } else {
        html.replace(i, 1, "&loz;");  // *
      }

    } else if ('-' == h_i) {  // put a longer - char
      if (' ' == html[i - 1] && ' ' == html[i + 1]) {
        html.replace(i - 1, 3, " &#x2013; ");
      }

    } else if ('/' == h_i) {
      // deal with /<
      if ('<' == html[i + 1]) {
        html.replace(i, 2, "Din ");
      }

      // deal with / <
      if ('<' == html[i + 2]) {
        html.replace(i, 3, "Din ");
      }

    } else if ('[' == h_i) {
      // deal with [<
      if ('<' == html[i + 1]) {
        html.replace(i, 2, "[Din ");
      }

      // deal with [ <
      if ('<' == html[i + 2]) {
        html.replace(i, 3, "[Din ");
      }

    } else if ('(' == h_i) {
      // deal with (<
      if ('<' == html[i + 1]) {
        html.replace(i, 2, "(Din ");
      }

      // deal with ( <
      if ('<' == html[i + 2]) {
        html.replace(i, 3, "(Din ");
      }

    } else if ('<' == h_i) {  // deal with <
      // do not process html close tags
      if ('/' != html[i + 1]) {
        html.replace(i, 1, "Provine din ");
      }
    }

    i++;
  } while (i < html.length());

  return html;
}

QString MainWindow::GetDictionaryName(const QString &shortName) {
  QString result;
  int row;

  for (row = 1; row < DICTIONARIES_COUNT; row++) {
    if (DICTIONARIES[row][0] == shortName) {
      result = QString("<a href=\"!surse\">%1</a>")
               .arg(QString::fromUtf8(DICTIONARIES[row][1]));
      break;
    }
  }

  // Take unknown dictionary name from XML, and don't add a link for it
  if (DICTIONARIES_COUNT == row) {
    result = shortName;
  }

  return result;
}

// Htmlize an entry for the w_browser.
QString MainWindow::HtmlizeBrowserEntry(const QString &dbRec) {
  int idx0 = 0, idx1;
  int dataLen = dbRec.length();

  for (idx1 = 0; idx1 < dataLen; idx1++) {
    if ('<' == dbRec[idx1]) {
      idx0 = idx1;
      continue;
    }

    if ('>' == dbRec[idx1]) {
      break;
    }
  }

  QString result = "<table cellpadding=0 cellspacing=3 width=\"100%\"><tr><td>";
  result += HtmlizeRawDefinition(dbRec.mid(idx1 + 1,
                                 dataLen - idx1 - 1).toAscii().data());
  result += "<br><font size=-1 color='#708090'>Sursa: ";
  result += GetDictionaryName(dbRec.mid(idx0 + 1, idx1 - idx0 - 1));
  result += "</font></td></tr></table>";
  return result;
}

// This slot is linked to bigger Font action (Alt + +)
void MainWindow::MakeFontBigger(void) {
  QFont newFont = font();
  int newsz = newFont.pointSize() + 1;

  if (newsz > 16) {
    newsz = 16;
  }

  newFont.setPointSize(newsz);
  setFont(newFont);
  return;
}

// This slot is linked to smaller Font action (Alt + -)
void MainWindow::MakeFontSmaller(void) {
  QFont newFont = font();
  int newsz = newFont.pointSize() - 1;

  if (newsz < 8) {
    newsz = 8;
  }

  newFont.setPointSize(newsz);
  setFont(newFont);
  return;
}

// This slot is linked to change Font action (Alt + F)
void MainWindow::ChangeFont(void) {
  bool ok;
  QFont newFont = QFontDialog::getFont(&ok, font(), this);

  if (ok) {
    if (newFont.pointSize() > 16) {
      newFont.setPointSize(16);

    } else if (newFont.pointSize() < 8) {
      newFont.setPointSize(8);
    }

    setFont(newFont);
  }

  return;
}

// Report word search findings
QString MainWindow::ReportResults(const QString &word,
                                  const QString &distinctWords,
                                  const QString &definitions,
                                  int stats[]) {
  QString buf, results =
    QString::fromUtf8("<table cellpadding=\"0\" cellspacing=\"1\" "
                      "width=\"100%\"><tr><td bgcolor=\"#eeeeee\">");

  // stats[0] distinct words #
  // stats[1] exact matches #
  // stats[2] definitions #
  // stats[3] dictionary entries #
  // stats[4] distinct dictionary words #
  // stats[5] words #

  if (0 == stats[0]) {
    // CASE 1: Nothing was found, info message is displayed.
    buf =
      QString::fromUtf8("Nici o definiție nu a fost găsită pentru «%1»."
                        "</td></tr></table>").arg(word);
    results += buf;

  } else if (1 == stats[0] && stats[1]) {
    // CASE 2: One distinct word was found, its definition[s] is[are] shown.
    if (0 != m_defaultDictionary && stats[2] < stats[1]) {
      if (stats[2] <= MAX_DEFINITIONS_DISPLAYED) {
        buf =
          QString::fromUtf8(" din toate sursele (%1 din %2)")
          .arg(stats[2])
          .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));

      } else {
        buf =
          QString::fromUtf8(" din toate sursele (mai mult de %1 din %2)")
          .arg(MAX_DEFINITIONS_DISPLAYED)
          .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));
      }
    }

    results +=
      QString::fromUtf8("%1 definiți%2%3 pentru «%4»:</td></tr><tr><td align="
                        "\"right\"><font size=-1 color='#686868'><a href="
                        "\"!abr\">Abrevieri</a></font></td></tr></table>")
      .arg(stats[1])
      .arg(1 == stats[1] ? "e" : "i")
      .arg(buf)
      .arg(word);

    if (stats[2] <= MAX_DEFINITIONS_DISPLAYED) {
      results += definitions;
    }

  } else if (stats[3] <= MAX_DEFINITIONS_DISPLAYED) {
    // CASE 3: Distinct words were found, user may chose one from a list
    buf =
      QString::fromUtf8(", %1 definiți%2 pentru %3 %4 fiind din %5")
      .arg(stats[3])
      .arg(1 == stats[3] ? "e" : "i")
      .arg(stats[4])
      .arg(1 == stats[4]
           ? QString::fromUtf8("cuvânt") : "cuvinte")
      .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));
    results +=
      QString::fromUtf8("Căutarea «%1» a găsit %2 definiți%3 pentru %4 %5 "
                        "din toate sursele%6.</td></tr></table>")
      .arg(word)
      .arg(stats[5])
      .arg(stats[5] == 1 ? "e" : "i")
      .arg(stats[0])
      .arg(1 == stats[0]
           ? QString::fromUtf8("cuvânt") : "cuvinte")
      .arg(0 == m_defaultDictionary ? QString::fromUtf8("") : buf);
    results += distinctWords + "|";

  } else if (stats[3] > MAX_DEFINITIONS_DISPLAYED) {
    // CASE 4: Too many words were found, a narrow search is required
    QString buf2 =
      QString::fromUtf8(" (%1 definiții din %2)")
      .arg(stats[3])
      .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));
    buf =
      QString::fromUtf8("%1 rezultate pentru «%2»%3.</td></tr></table><p>"
                        "<font color=\"#993300\">Numărul maxim de definiții "
                        "care pot fi afișate este %4.<br>Folosiți un "
                        "criteriu de căutare mai selectiv.</font></p>")
      .arg(stats[5])
      .arg(word)
      .arg(0 == m_defaultDictionary ? QString::fromUtf8("") : buf2)
      .arg(MAX_DEFINITIONS_DISPLAYED);
    results += buf;
  }

  return results;
}

// Called from the constructor.
void MainWindow::DbStartIndexing(void) {
  ClearBrowser();
  w_historyListBox->clear();

  // DbSetup returns true if an update is required
  if (DbSetup()) {
    GenericUpdate();

  } else {
    // Build an in-memory word index, asynchronously.
    m_pThread->start();
    SetBrowserHtml(ProcessCommands("!pp"));
  }

  SetFocusToSearchBox();
  return;
}

// Return true if a database update is required, otherwise return false
bool MainWindow::DbSetup(void) {
  int iarg = 1;

  // Jump over potential command line options.
  while (qApp->argc() > iarg && qApp->argv()[iarg][0] == '-') {
    iarg++;
  }

  QDir dbd;  // Current directory.
  dbd.makeAbsolute();

  // Get full path for db file.
  if (qApp->argc() > iarg) {
    qstrncpy(m_dbFileName,
             QDir::cleanPath(dbd.filePath(qApp->argv()[iarg])).toAscii(), 128);

  } else {
    qstrncpy(m_dbFileName, dbd.filePath(DB_FILE_NAME).toAscii(), 128);
  }

  m_pThread = new CIndexThread(this);
  LogInfo(QString("Info: Database file %1").arg(m_dbFileName));

  // Find last update time under <> key in the database.
  bool no_database = false;

  if (!DbOpenRead()) {
    // Probably a new installation.
    no_database = true;
    LogInfo("Info: The program will create a new database.");

  } else {
    bool ok;
    char *dbLastUpdateTime;
    dbLastUpdateTime = DbGet(LAST_UPDATE_TIME_KEY, &ok);

    if (ok) {
      m_lastUpdateTime = QByteArray(dbLastUpdateTime).toInt();
      DbFree(dbLastUpdateTime);

    } else {
      LogInfo("Error: No timestamp was found in the database.");
    }
  }

  DbClose();
  return (m_autoUpdate || no_database);
}

// Save XML data. Return the number of chars processed.
int MainWindow::DbSaveXMLData(const QString &xmlData,
                              UpdateErrors *updateError,
                              bool *abortHttp) {
  // This is the last update time advertised by the server at the beginning
  // of the transmition. It is used only when all data was received okay.
  // Otherwise, the individual time in each entry is taken.
  static time_t lutAdvertised = 0;

  QChar nextChar1, nextChar2;
  int   lastIdx = 0, idx = 0;
  int   oIdx[6] = {0, 0, 0, 0, 0, 0};
  int   cIdx[6] = {0, 0, 0, 0, 0, 0};
  QString dbNewWord, dbNewRec, entryId, entryWord, entryTimeStamp;

  *updateError = UPDATE_ERR_NOT_DONE;
  int dataLen  = xmlData.length();

  while (idx < dataLen) {
    // look only for tags
    if ('<' != xmlData[idx]) {
      idx++;
      continue;
    }

    nextChar1 = xmlData[idx + 1];
    nextChar2 = xmlData[idx + 2];

    // Check if we haven't seen an entry yet.
    if (0 == oIdx[0]) {
      // Test for the update timestamp.
      if ('T' == nextChar1) {
        idx += 11;
        oIdx[2] = idx;
        continue;

      } else if ('/' == nextChar1 && 'T' == nextChar2) {
        cIdx[2] = idx;
        idx += 12;
        lutAdvertised = xmlData.mid(oIdx[2], cIdx[2] - oIdx[2]).toInt();
        continue;

      } else if ('N' == nextChar1 && 'u' == nextChar2) {
        idx += 12;
        oIdx[5] = idx;
        continue;

      } else if ('/' == nextChar1
                 && 'N' == nextChar2 && 'u' == xmlData[idx + 3]) {
        cIdx[5] = idx;
        m_newDefCountPending = xmlData.mid(oIdx[5], cIdx[5] - oIdx[5]).toInt();

        if (m_newDefCountPending) {
          // Open the database.
          if (!DbOpenWrite()) {
            *updateError = UPDATE_ERR_DB_WRITE;
            *abortHttp = true;
            return 0;
          }
        }

        idx += 13;
        continue;
      }
    }

    if ('/' == nextChar1  &&  'D' == nextChar2  &&  'i' == xmlData[idx + 3]) {
      if (lutAdvertised > 0) {
        m_lastUpdateTimePending = lutAdvertised;
      }

      DbClose();
      *updateError = UPDATE_ERR_NO_ERROR;
      *abortHttp = true;
      idx += 13;

    } else if ('T' == nextChar1) {
      idx += 11;
      oIdx[2] = idx;

    } else if ('/' == nextChar1 && 'T' == nextChar2) {
      cIdx[2] = idx;
      idx += 12;

    } else if ('S' == nextChar1) {
      idx += 8;
      oIdx[3] = idx;

    } else if ('/' == nextChar1 && 'S' == nextChar2) {
      cIdx[3] = idx;
      idx += 9;

    } else if ('D' == nextChar1 && 'e' == nextChar2) {
      idx += 12;
      oIdx[4] = idx;

    } else if ('/' == nextChar1
               && 'D' == nextChar2 && 'e' == xmlData[idx + 3]) {
      cIdx[4] = idx;
      idx += 13;

    } else if ('N' == nextChar1 && 'a' == nextChar2) {
      idx += 6;
      oIdx[1] = idx;

    } else if ('/' == nextChar1
               && 'N' == nextChar2 && 'a' == xmlData[idx + 3]) {
      cIdx[1] = idx;
      idx += 7;

    } else if ('I' == nextChar1 && 'd' == nextChar2) {
      idx += 4;
      oIdx[0] = idx;

    } else if ('/' == nextChar1
               && 'I' == nextChar2 && 'd' == xmlData[idx + 3]) {
      cIdx[0] = idx;
      idx += 5;

      // have a record, update the db

    } else if ('/' == nextChar1
               && 'E' == nextChar2 && 'n' == xmlData[idx + 3]) {
      if (-1 == m_newDefCountPending) {
        *updateError = UPDATE_ERR_CONTENT;
        *abortHttp = true;
        return 0;
      }

      if (!DbActive()) {
        LogInfo("Error: Database is not open during update.");
        *updateError = UPDATE_ERR_DB_WRITE;
        *abortHttp = true;
        return 0;
      }

      entryId        = xmlData.mid(oIdx[0], cIdx[0] - oIdx[0]);
      entryWord      = xmlData.mid(oIdx[1], cIdx[1] - oIdx[1]);
      entryTimeStamp = xmlData.mid(oIdx[2], cIdx[2] - oIdx[2]);

      dbNewRec = QString("%1<%2>%3")
                 .arg(entryId)
                 .arg(xmlData.mid(oIdx[3], cIdx[3] - oIdx[3]))
                 .arg(xmlData.mid(oIdx[4], cIdx[4] - oIdx[4]));

      bool  ok = true, dbError = false, sameKey;
      char *dupe;
      int   ij = 0, keySz;
      dbNewWord = entryWord;

      while (DbKeyExists(dbNewWord.toAscii().data())) {
        // Check whether this entry word exists with the same Id.
        dupe = DbGet(dbNewWord.toAscii().data(), &ok);

        if (ok) {
          sameKey = true;
          keySz = entryId.length();

          if ('<' != dupe[keySz]) {
            sameKey = false;

          } else {
            for (int k = 0; k < keySz; k++) {
              if (entryId[k] != dupe[k]) {
                sameKey = false;
                break;
              }
            }
          }

          DbFree(dupe);

          if (sameKey) {
            break;
          }

        } else {
          LogInfo(QString("Error: Updated entry word \"%1\" exists, "
                          "but cannot be retrieved").arg(entryWord));
          dbError = true;
        }

        dbNewWord = QString("%1%2").arg(entryWord).arg(ij);
        ij++;
      }

      if (dbError) {
        *updateError = UPDATE_ERR_DB_READ;
        *abortHttp = true;
        return 0;
      }

      if (!DbPut(dbNewWord.toAscii().data(), dbNewRec.toAscii().data())) {
        *updateError = UPDATE_ERR_DB_WRITE;
        *abortHttp = true;
        return 0;
      }

      m_lastUpdateTimePending = entryTimeStamp.toInt() - 1;
      m_newDefCount++;

      if (m_newDefCount + MAX_NEW_DEFS_TO_DISPLAY > m_newDefCountPending) {
        m_lstNewDefinitions.push_front(dbNewRec);

        if (m_lstNewDefinitions.size() > MAX_NEW_DEFS_TO_DISPLAY) {
          m_lstNewDefinitions.pop_back();
        }
      }

      idx += 8;
      lastIdx = idx;

    } else {
      idx++;
    }
  }

  return lastIdx;
}

QString MainWindow::DbResults(const QString &word) {
  if (!DbOpenRead()) {
    DbClose();
    return
      QString::fromUtf8("<font color='#993300'>Eroare: Programul nu poate "
                        "citi din baza de date.</font><br> Verificați "
                        "existența bazei de date (%1). Dacă există, s-ar "
                        "putea să fie în curs de actualizare de către altă "
                        "instanță a programului, sau s-ar putea să fie "
                        "coruptă.<br>")
      .arg(m_dbFileName);
  }

  bool duplicates;
  int ij;
  QStringList lstWords;
  QStringList::iterator it;
  QString awij, lWord(word);

  QString buf = QString::fromAscii("^(") + lWord;
  buf.replace('*', "[ \\-a-z]*");
  buf.replace('?', "[ \\-a-z]");
  buf += QString::fromAscii(")$");

  // Case insensitive word search - need to get DE words (e.g. Adam)
  QRegExp rx = QRegExp(buf, Qt::CaseInsensitive, QRegExp::RegExp);
  rx.setMinimal(true);

  // Find all matching words by searching in m_lstAllWords.
  for (it = m_lstAllWords.begin(); it != m_lstAllWords.end(); it++) {
    buf = *it;

    if (-1 != buf.indexOf(rx)) {
      lstWords.append(buf);
      // Check for other word entries for this word.
      // They are stored in the database with keys word0, word1, etc
      ij = 0;

      do {
        awij = QString("%1%2").arg(buf).arg(ij);
        duplicates = DbKeyExists(awij.toAscii().data());

        if (duplicates) {
          lstWords.append(awij);
        }

        ij++;
      } while (duplicates);
    }
  }

  int stats[] = {0, 0, 0, 0, 0, 0};
  stats[1] = lstWords.size();  // exact matches counter

  // If no exact matches, try to find close matches.
  if (!stats[1]) {
    int  editDistance;

    for (it = m_lstAllWords.begin(); it != m_lstAllWords.end(); it++) {
      buf = *it;

      if (lWord.length() + 2 <= buf.length()
          || lWord.length() >= buf.length() + 2) {
        continue;
      }

      editDistance = LevenshteinDistance(lWord, buf, 2);

      if (editDistance < 2) {
        lstWords.append(buf);

        if (lstWords.size() > MAX_DEFINITIONS_DISPLAYED) {
          break;
        }
      }
    }
  }

  // Count distinct words in lstWords
  QSet<QString> setWords;

  for (it = lstWords.begin(); it != lstWords.end(); it++) {
    buf = *it;
    rx = QRegExp("[0-9]");

    if (-1 == buf.indexOf(rx)) {
      setWords.insert(buf.toLower());
    }
  }

  // Stats update
  stats[0] = setWords.size();  // distinct words counter
  stats[5] = lstWords.size();  // words counter

  // Fetch definitions only for stats[0] = 1 and also update stats
  QString definitions("");
  QStringList lstDictWords;

  if (stats[5]) {
    bool ok;
    char *dbRec;
    QString dbRawRec, dictionary, htmlEntry;

    if (0 == m_defaultDictionary) {
      dictionary = QString("<");

    } else {
      dictionary = QString("<%1>").arg(DICTIONARIES[m_defaultDictionary][0]);
    }

    for (it = lstWords.begin();  // stats[2] -> definitions counter
         it != lstWords.end() && stats[2] <= MAX_DEFINITIONS_DISPLAYED;
         it++) {
      dbRec = DbGet((*it).toAscii().data(), &ok);

      if (ok) {
        dbRawRec = QString(dbRec);
        DbFree(dbRec);

        if (-1 == dbRawRec.indexOf(dictionary)) {
          continue;

        } else {
          lstDictWords.append(*it);
          stats[3]++;  // dictionary entries counter
        }

        if (1 == stats[0]) {  // distinct words counter
          htmlEntry = HtmlizeBrowserEntry(dbRawRec);
        }

      } else {
        htmlEntry =
          QString::fromUtf8("<p><font color='#993300'>Eroare internă: "
                            "Programul nu a găsit cuvântul %1 în "
                            "baza de date.</font></p>")
          .arg(*it);
      }

      if (1 == stats[0]) {  // distinct words counter
        definitions += htmlEntry;
        stats[2]++;  // definitions counter
      }
    }
  }

  DbClose();

  // Add all distinct words to distinctWords string, and also update stats.
  QString distinctWords("");

  if (stats[3] <= MAX_DEFINITIONS_DISPLAYED) {  // dictionary entries
    distinctWords = "<br><br>";
    rx = QRegExp("[0-9]");

    if (0 == m_defaultDictionary) {
      lstWords.sort();

      for (it = lstWords.begin(); it != lstWords.end(); it++) {
        buf = *it;

        if (-1 == buf.indexOf(rx)) {
          distinctWords += "| <a href=\"" + buf + "\">" + buf + "</a> ";
        }
      }

    } else {
      lstDictWords.sort();
      QString prevBuf("");

      for (it = lstDictWords.begin(); it != lstDictWords.end(); it++) {
        buf = *it;
        buf.replace(rx, "");

        // Suffixless word is in the list, OR not in the list but word_ij is
        if (buf == (*it)
            || (-1 == lstDictWords.indexOf(buf) && buf != prevBuf)) {
          stats[4]++;  // distinct dictionary words counter
          distinctWords += "| <a href=\"" + buf + "\">" + buf + "</a> ";
        }

        prevBuf = buf;
      }
    }
  }

  return ReportResults(word, distinctWords, definitions, stats);
}

// This function runs in its own thread
void MainWindow::DbCreateIndex(void) {
  if (!DbOpenRead()) {
    m_updateError = UPDATE_ERR_DB_READ;
    return;
  }

  // initialize the iterator
  if (!dpiterinit(m_dbHandle)) {
    LogInfo(QString("Error: %1 - Dixit cannot initialize the db iterator.")
            .arg(dperrmsg(dpecode)));
    m_updateError = UPDATE_ERR_DB_READ;
    return;
  }

  char *key;
  int keySz, count = 0;

  m_lstAllWords.clear();

  // scan with the iterator to get all keys.
  while (NULL != (key = dpiternext(m_dbHandle, &keySz))) {
    count++;

    // Store each word only once in the m_lstAllWords list, ie skip duplicates.
    // (duplicate words have a number at the end in the database)
    if (!(key[keySz - 1] >= '0' && key[keySz - 1] <= '9')) {
      m_lstAllWords.append(key);
    }

    free(key);
  }

  keySz = 1;

  if (DbKeyExists(CORRELATION_KEY)) {
    bool ok = true;
    key = DbGet(CORRELATION_KEY, &ok);

    if (ok) {
      keySz = QByteArray(key).toInt();
      free(key);
    }
  }

  DbClose();
  m_totalDefinitions = count - keySz;
  return;
}

bool MainWindow::DbOpenRead(void) {
  if (!(m_dbHandle = dpopen(m_dbFileName, DP_OREADER, -1))) {
    LogInfo(QString("Error: %1 - Dixit cannot open database for reading.")
            .arg(dperrmsg(dpecode)));
    return false;
  }

  return true;
}

bool MainWindow::DbOpenWrite(void) {
  if (!(m_dbHandle = dpopen(m_dbFileName, DP_OWRITER | DP_OCREAT, -1))) {
    LogInfo(QString("Error: %1 - Dixit cannot open database for writing.")
            .arg(dperrmsg(dpecode)));
    return false;
  }

  return true;
}

bool MainWindow::DbClose(void) {
  if (m_dbHandle) {
    if (!dpclose(m_dbHandle)) {
      m_dbHandle = NULL;
      LogInfo(QString("Error: %1 - Dixit cannot close database.")
              .arg(dperrmsg(dpecode)));
      return false;
    }
  }

  m_dbHandle = NULL;
  return true;
}

bool MainWindow::DbKeyExists(const char *key) {
  return (-1 != dpvsiz(m_dbHandle, key, -1));
}

bool MainWindow::DbPut(const char *key,
                       const char *value) {
  if (!dpput(m_dbHandle, key, -1, value, -1, DP_DOVER)) {
    LogInfo(QString("Error: %1 - Dixit cannot save record for key \"%2\".")
            .arg(dperrmsg(dpecode))
            .arg(key));
    return false;
  }

  return true;
}

// Note: dpget() allocates memory that must be released by DbFree
char *MainWindow::DbGet(const char *key,
                        bool *ok) {
  char *value = dpget(m_dbHandle, key, -1, 0, -1, NULL);

  if (value) {
    *ok = true;

  } else {
    *ok = false;
    LogInfo(QString("Error: %1 - Dixit cannot get the value for key \"%2\".")
            .arg(dperrmsg(dpecode))
            .arg(key));
  }

  return value;
}

// Note: Frees memory allocated by DbGet.
void MainWindow::DbFree(char *item) {
  if (item) {
    free(item);
  }

  return;
}

// Checks if database is active
bool MainWindow::DbActive(void) {
  return (m_dbHandle ? true : false);
}

// END OF FILE

